	var selRange;
    var wordHighlightApplier;

    var isScrolling = false;
    // Ordinal values of ReaderCursorPositionPageSetting enum in Java
    var CURSOR_PAGE_POSITION_ON_PAGE = 0;
    var CURSOR_PAGE_POSITION_CENTERED = 1;
    var manualScroll = false;
    var cursorIsVisible = true;
    var isFullScreen = false;
    var initialCursorShowRequested = false;

    var SPOKEN_LINE_ELEMENT_ID = "vd-spoken-line";
    var SPOKEN_LINE_CSS_CLASS = "voice-dream-spoken-line";
    var SPOKEN_WORD_CSS_CLASS = "voice-dream-spoken-word";

    var SPINNER_DIV_ID = "center";
    var INIT_SPINNER_DIV_ID = "center_init";

    function getLineHighlightElement()
    {
        //add our line highlight div if not present already
        var lineHighlightElement = document.getElementById(SPOKEN_LINE_ELEMENT_ID);
        if( lineHighlightElement == null )
        {
            // append our line highlight div, floating
            var newDiv = $("<div>", {id:SPOKEN_LINE_ELEMENT_ID, class:SPOKEN_LINE_CSS_CLASS});
            $( "body" ).append(newDiv);

            lineHighlightElement = document.getElementById(SPOKEN_LINE_ELEMENT_ID);
        }
        return lineHighlightElement;
    }



    var showingSpokenWord = true;
    var showingSpokenLine = true;

    function highlightCursor(start, end, elementIndex, cursorPagePositionSetting, showSpokenWord, showSpokenLine)
    {
        showingSpokenWord = showSpokenWord;
        showingSpokenLine = showSpokenLine;

        if( wordHighlightApplier == null )
        {
            wordHighlightApplier = rangy.createClassApplier(SPOKEN_WORD_CSS_CLASS,
                { elementTagName: "vdword",
                  normalize: true
                }
            );
        }
        if( selRange == null )
        {
            selRange = rangy.createRange();
        }
        else
        {
            wordHighlightApplier.undoToRange(selRange);
        }

        var highlightNode = getNodeForElementIndex(elementIndex);
        if( highlightNode == null ) return;

        selRange.selectCharacters(highlightNode, start, end);
        // Removing this (1-Jun-2015) as it appears to cause
        // some weird behaviour with end of sentence words
        // which have non-alphanumerics post-pended eg ", or !" etc.
        // see pride & prejudice - particularly with acapela voices
        //selRange.expand("word");
        wordHighlightApplier.applyToRange(selRange);

        drawLineHighlight();
        var theWordHighlight = getWordHighlight();


        if( !showSpokenWord )
        {
            //$(theWordHighlight).css('backgroundColor', $.Color({ alpha: 0.0 }));
            $(theWordHighlight).removeClass(SPOKEN_WORD_CSS_CLASS);
        }


        if( cursorIsVisible )
        {
            if( cursorPagePositionSetting == CURSOR_PAGE_POSITION_ON_PAGE )
            {
                if(!isElementInViewport(theWordHighlight) && !isScrolling)
                {
                    scrollElementToTop(theWordHighlight);
                }
            }
            else if( cursorPagePositionSetting == CURSOR_PAGE_POSITION_CENTERED )
            {
                if( !isScrolling )
                {
                    scrollElementToMiddle(theWordHighlight);
                }
            }
        }

        cursorIsVisible = isElementInViewport(theWordHighlight);

        if( cursorIsVisible )
        {
            notifyCursorVisibility();
        }

    }

    function getWordHighlight()
    {
        var wordHighlights = document.getElementsByTagName("vdword");
        return wordHighlights[wordHighlights.length - 1];
    }

    function drawLineHighlight()
    {
        var lineHighlight = getLineHighlightElement();
        var theWordHighlight = getWordHighlight();

        if( showingSpokenLine )
        {
            var wordRect = theWordHighlight.getBoundingClientRect();
            var scrollTop = $(window).scrollTop();

            var lhTop = wordRect.top + scrollTop;

            if( wordRect.height == 0 )
            {
                var fontsize = $("body").css('font-size');

                lineHighlight.style.top = lhTop+5  + 'px';
                lineHighlight.style.height = fontsize + 'px';
            }
            else
            {
                lineHighlight.style.top = lhTop  + 'px';
                lineHighlight.style.height = wordRect.height + 'px';
            }

        }
        else
        {
            lineHighlight.style.top = '0px';
            lineHighlight.style.height = '0px';
        }
    }

    function setCursorVisible(visibility, cursorPagePositionSetting, fullScreen)
    {
        cursorIsVisible = visibility;
        isFullScreen = fullScreen;
        if( cursorIsVisible )
        {
            scrollCursorIntoView(cursorPagePositionSetting);
        }
    }


    function isElementInViewport (el) {

        //special bonus for those using jQuery
        if (typeof jQuery === "function" && el instanceof jQuery)
        {
            el = el[0];
        }

        var rect = el.getBoundingClientRect();


        var elementInViewport = (
        rect.top >= 0 &&
        rect.left >= 0 &&
        rect.bottom <= (window.innerHeight || document.documentElement.clientHeight) && /*or $(window).height() */
        rect.right <= (window.innerWidth || document.documentElement.clientWidth) /*or $(window).width() */
        );
        return elementInViewport;
    }

    function getIndexOfVisibleReadableElement()
    {
        var height = (window.innerHeight || document.documentElement.clientHeight);

        // iterate from 0 - height in increments of 20 to find the first readable element
        var xLocation = (window.innerWidth || document.documentElement.clientWidth) / 2;
        var yLocation = 0;

        var vdReadableElemAttribute = "[" +  VD_READABLE_ELEMENT_INDEX_ATTRIBUTE + "]";
        while(yLocation < height)
        {
            var elemAtPoint = document.elementFromPoint(xLocation, yLocation);
            if( elemAtPoint != null )
            {
                if( elemAtPoint.hasAttribute(VD_READABLE_ELEMENT_INDEX_ATTRIBUTE))
                {
                    var elemIndexStr = elemAtPoint.getAttribute(VD_READABLE_ELEMENT_INDEX_ATTRIBUTE);
                    var elemIndex = parseInt(elemIndexStr);
                    return elemIndex;

                }
            }

            yLocation += 20;
        }

        return -1;
    }

    function clearLineHighlight()
    {
        $("#"+SPOKEN_LINE_ELEMENT_ID).remove();
    }

    function clearWordHighlight()
    {
        wordHighlightApplier.undoToRange(selRange);
    }

    function getNodeForElementIndex(elementIndex)
    {
        var elementSelector = "[" + VD_READABLE_ELEMENT_INDEX_ATTRIBUTE + "='" + elementIndex + "']";
        var result = document.querySelector(elementSelector);
        return result;
    }

    function selectText(start, end, elementIndex)
    {
        var containerEl = getNodeForElementIndex(elementIndex);

        var charIndex = 0, range = document.createRange();
        range.setStart(containerEl, 0);
        range.collapse(true);
        var nodeStack = [containerEl], node, foundStart = false, stop = false;

        while (!stop && (node = nodeStack.pop()))
        {
            if (node.nodeType == 3)
            {
                var nextCharIndex = charIndex + node.length;
                if (!foundStart && start >= charIndex && start <= nextCharIndex)
                {
                    range.setStart(node, start - charIndex);
                    foundStart = true;
                }
                if (foundStart && end >= charIndex && end <= nextCharIndex)
                {
                    range.setEnd(node, end - charIndex);
                    stop = true;
                }
                charIndex = nextCharIndex;
            }
            else
            {
                var i = node.childNodes.length;
                while (i--)
                {
                    nodeStack.push(node.childNodes[i]);
                }
            }
        }

        var sel = window.getSelection();
        sel.removeAllRanges();
        sel.addRange(range);

    }



