/**
 * Created by michael on 12/9/14.
 */

// Ordinal values of ReaderCursorPositionPageSetting enum in Java
var HIGHLIGHT_STYLE_ON_PAGE = 0;
var HIGHLIGHT_STYLE_IN_MARGIN = 1;

var nextHighlightId = 0;
var HIGHLIGHT_ID_PREFIX = "vdid";

var VOICE_DREAM_PAGE_HIGHLIGHT_CLASS = "voice-dream-highlight";
var VOICE_DREAM_SIDEBAR_HIGHLIGHT_CLASS = "voice-dream-highlight-sidebar";
var VOICE_DREAM_MARGIN_HIGHLIGHT_CLASS =  "voice-dream-margin-highlight";
var VOICE_DREAM_NOTE_HIGHLIGHT_CLASS = "voice-dream-note-highlight";
var VOICE_DREAM_MARK_ID_ATTR = "vd-mark-id";
var VD_READABLE_ELEMENT_ATTRIBUTE = "_vdtxt";
var VD_READABLE_ELEMENT_INDEX_ATTRIBUTE = "_vdti";

var imageElementSingleTapDetected = false;

function getNextHighlightId()
{
    var highlightIdText =  HIGHLIGHT_ID_PREFIX + nextHighlightId;
    nextHighlightId++;
    return highlightIdText;
}


function drawBookmarkForTextRange(startNodeIndex, startOffset, endOffset, scrollToTop)
{
    // setup the bookmark img element
    var bookmarkId = getNextHighlightId();
    var bookmark = document.createElement("img");
    bookmark.id = bookmarkId;
    bookmark.src="png/bookmark.png";
    bookmark.className="voice-dream-bookmark-highlight";
    $( "body" ).append(bookmark);

    // setup a rangy range encompassing the start of the bookmark selection and stick a
    // class selection around it
    var selHighlightId = getNextHighlightId();
    var selHighlightApplier = rangy.createClassApplier(VOICE_DREAM_PAGE_HIGHLIGHT_CLASS,
        { elementTagName: "vdhighlight",
            elementProperties: {
                id: selHighlightId
            }
        }
    );
    var startNode = getNodeForElementIndex(startNodeIndex);

    var selRange = rangy.createRange();
    selRange.selectCharacters(startNode, startOffset, endOffset);
    selHighlightApplier.applyToRange(selRange);

    // Now get the selection range and use its location to position the bookmark image/icon
    var insertedMarkerRange = document.getElementById(selHighlightId);
    var markerClientRect = insertedMarkerRange.getBoundingClientRect();

    // given the selection range client rect, figure out where the top
    // of the bookmark button should be located
    var spanTop = markerClientRect.top + $(window).scrollTop();

    // and find our newly added bookmark button so we can style/position it
    var bookmarkButton = document.getElementById(bookmarkId);
    bookmarkButton.style.top = spanTop  + 'px';
    bookmarkButton.style.height='12px';
    bookmarkButton.style.width='12px';

    if( scrollToTop )
    {
        scrollElementToTop(bookmarkButton);
    }


    // finally remove our selection highlight since we no longer need it
    selHighlightApplier.undoToRange(selRange);
}

function highlightTextRange(startNodeIndex, startOffset, endNodeIndex, endOffset, highlightStyle, withNote, markId)
{

    // Setup the css class for highlights. Note only the on-page highlight actually has css
    // properties here, since for in-margin highlights we dont want to display anything directly
    // but instead just use the span class to define the size of the sidebar highlight
    var highlightCssClass = highlightStyle == HIGHLIGHT_STYLE_ON_PAGE ? VOICE_DREAM_PAGE_HIGHLIGHT_CLASS
                                                                      : VOICE_DREAM_SIDEBAR_HIGHLIGHT_CLASS;

    var noteIconId = "";
    if( withNote )
    {
        // setup a note icon image
        noteIconId = getNextHighlightId();
        var noteIcon = document.createElement("img");
        noteIcon.id = noteIconId;
        noteIcon.src = "png/yellow_postit.png";
        noteIcon.className = VOICE_DREAM_NOTE_HIGHLIGHT_CLASS;
        $("body").append(noteIcon);
    }

    var selIndex = startNodeIndex;
    var topmostHighlightElementId = "";
    var bottomMostHighlightElementId = "";

    do
    {
        var selRange = rangy.createRange();

        var selHighlightId = getNextHighlightId();
        var selHighlightApplier = rangy.createClassApplier(highlightCssClass,
            { elementTagName: "vdhighlight",
                elementProperties: {
                    id: selHighlightId
                }
            }
        );

        var vdNode = getNodeForElementIndex(selIndex);
        if( selIndex == startNodeIndex )
        {
            topmostHighlightElementId = selHighlightId;

            if( endNodeIndex == startNodeIndex )
            {
                // pretty simple - just select the characters between the offsets within this node
                selRange.selectCharacters(vdNode, startOffset, endOffset);
                bottomMostHighlightElementId = selHighlightId;
            }
            else
            {
                // on the first node of a multi-node selection, so
                // select everything from the start offset in this node to the end of this node
                // figure out the range of this vd node (so we know how many
                // chars to select to cover to the end of the node
                var vdNodeAttr = vdNode.getAttribute(VD_READABLE_ELEMENT_ATTRIBUTE);
                var attrVals = vdNodeAttr.split(":");
                var nodeLength = parseInt(attrVals[1]);

                selRange.selectCharacters(vdNode, startOffset, nodeLength);
            }
        }
        else if( selIndex == endNodeIndex )
        {
            // at the last node of a multi-node selection, so
            // select everything from the start of this node to the end offset
            selRange.selectCharacters(vdNode, 0, endOffset);
            bottomMostHighlightElementId = selHighlightId;
        }
        else
        {
            // select the entire node content
            selRange.selectNode(vdNode);
        }

        // highlight the current range with a class applier
        selHighlightApplier.applyToRange(selRange);
        var highlightedElement = document.getElementById(selHighlightId);
        highlightedElement.setAttribute(VOICE_DREAM_MARK_ID_ATTR,markId);

        selIndex++;
    }
    while( selIndex <= endNodeIndex );

    // if we're highlighting on-page then the above selection will have the correct highlighting
    // css class; however for a margin highlight we need to draw the sidebar highlight ourselves
    // by creating a floating div corresponding to the coordinates of the selection range
    if( highlightStyle == HIGHLIGHT_STYLE_IN_MARGIN )
    {
        var sidebarHighlightId = getNextHighlightId();
        var newSidebarHighlightDiv = $("<div>", {id:sidebarHighlightId, class:VOICE_DREAM_MARGIN_HIGHLIGHT_CLASS});
        $( "body" ).append(newSidebarHighlightDiv);

        // find the in-page elements for the top/bottom of the highlight (may be
        // the same for a small highlight)
        var startHighlightSpanElement = document.getElementById(topmostHighlightElementId);
        var endHighlightSpanElement = document.getElementById(bottomMostHighlightElementId);

        // and find our newly added sidebar div
        var sidebarHighlight = document.getElementById(sidebarHighlightId);

        // now use its coordinates to set the coordinates on our new div
        var startHighlightSpanRect = startHighlightSpanElement.getBoundingClientRect();
        var endHighlightSpanRect = endHighlightSpanElement.getBoundingClientRect();

        var combinedSpanHeight = endHighlightSpanRect.bottom - startHighlightSpanRect.top;

        var scrollTop = $(window).scrollTop();

        var spanTop = startHighlightSpanRect.top + scrollTop;
        sidebarHighlight.style.top = spanTop  + 'px';
        sidebarHighlight.style.height = combinedSpanHeight + 'px';

        if( withNote )
        {
            var noteImgElement = document.getElementById(noteIconId);
            noteImgElement.style.top = spanTop  + 'px';
            noteImgElement.style.right = "5px";

            setupImageElementLongPressNotifier(noteImgElement);
        }
    }
    else if( withNote )
    {
        // In-page note icon, position to the rightmost point of the top of the highlight span
        var startHighlightSpanElement = document.getElementById(topmostHighlightElementId);
        var startHighlightSpanRect = startHighlightSpanElement.getBoundingClientRect();

        var spanTop = startHighlightSpanRect.top + $(window).scrollTop();
        var noteImgElement = document.getElementById(noteIconId);
        noteImgElement.style.top = (spanTop - 15)  + 'px';
        noteImgElement.style.left = (startHighlightSpanRect.right - 5) + 'px';

        setupImageElementLongPressNotifier(noteImgElement);
    }
}

function clearAllHighlights()
{
    $("."+VOICE_DREAM_PAGE_HIGHLIGHT_CLASS).contents().unwrap();
    $("."+VOICE_DREAM_SIDEBAR_HIGHLIGHT_CLASS).contents().unwrap();
    $("."+VOICE_DREAM_MARGIN_HIGHLIGHT_CLASS).remove();
    $("."+VOICE_DREAM_NOTE_HIGHLIGHT_CLASS).remove();
}

function getTextSelectionRangeForHighlight()
{
    var rsel = rangy.getSelection();
    if( rsel == null || rsel.rangeCount == 0) return;

    var selectionRange = rsel.getRangeAt(0);

    var rangeStart = getVoiceDreamElementIndexAndOffsetForSelectionBoundary(selectionRange.startContainer, selectionRange.startOffset);
    var rangeEnd = getVoiceDreamElementIndexAndOffsetForSelectionBoundary(selectionRange.endContainer, selectionRange.endOffset);

    // and notify the app of the double tap location details
    window.SynchJS.notifyTextSelectionRangeForHighlight(rangeStart.index, rangeStart.offset, rangeEnd.index, rangeEnd.offset);

}

/**
 * Text selection range for bookmarks is pretty much like that for a regular highlight except the end of the
 * range extends to cover the remainder of the container, not just to the end of the highlight/selection boundary
 */
function getTextSelectionRangeForBookmark()
{
    var rsel = rangy.getSelection();
    if( rsel == null || rsel.rangeCount == 0) return;

    var selectionRange = rsel.getRangeAt(0);

    var rangeStart = getVoiceDreamElementIndexAndOffsetForSelectionBoundary(selectionRange.startContainer, selectionRange.startOffset);
    var rangeEnd = getVoiceDreamElementIndexAndOffsetForSelectionBoundary(selectionRange.endContainer, selectionRange.endOffset);

    // and notify the app of the text range
    window.SynchJS.notifyTextSelectionRangeForBookmark(rangeStart.index, rangeStart.offset, rangeEnd.index);

}

function getTextSelectionRangeForNote(isForExistingNote, markIdOfExistingNote)
{
    var rsel = rangy.getSelection();
    if( rsel == null || rsel.rangeCount == 0) return;

    var selectionRange = rsel.getRangeAt(0);

    var rangeStart = getVoiceDreamElementIndexAndOffsetForSelectionBoundary(selectionRange.startContainer, selectionRange.startOffset);
    var rangeEnd = getVoiceDreamElementIndexAndOffsetForSelectionBoundary(selectionRange.endContainer, selectionRange.endOffset);

    // and notify the app of the double tap location details
    window.SynchJS.notifyTextSelectionRangeForNote(rangeStart.index, rangeStart.offset, rangeEnd.index, rangeEnd.offset, isForExistingNote, markIdOfExistingNote);

}

/**
 * As per the above bookmark function but instead of basing it on user selection we're just taking it from
 * the current word cursor
 */
function getTextSelectionRangeForBookmarkByCurrentWord()
{
    var wordHighlights = document.getElementsByTagName("vdword");
    if( wordHighlights == null ) return;

    var range = rangy.createRange();
    range.selectNodeContents(wordHighlights[0]);

    var rangeStart = getVoiceDreamElementIndexAndOffsetForSelectionBoundary(range.startContainer, range.startOffset);
    var rangeEnd = getVoiceDreamElementIndexAndOffsetForSelectionBoundary(range.endContainer, range.endOffset);

    // and notify the app of the double tap location details
    window.SynchJS.notifyTextSelectionRangeForBookmark(rangeStart.index, rangeStart.offset, rangeEnd.index);

}



function getVoiceDreamElementIndexAndOffsetForSelectionBoundary(boundaryContainer, boundaryOffsetInContainer)
{
    var vdboundaryContainingNode = boundaryContainer;
    while( vdboundaryContainingNode != null )
    {
        if(vdboundaryContainingNode.hasAttribute && vdboundaryContainingNode.hasAttribute(VD_READABLE_ELEMENT_ATTRIBUTE))
        {
            break;
        }
        else
        {
            vdboundaryContainingNode = vdboundaryContainingNode.parentNode;
        }
    }

    if( vdboundaryContainingNode == null ) return;

    // get the index of our voicedream node that holds this boundary node
    //var boundaryNodeIndex = $(VD_cssSelector).index(vdboundaryContainingNode);
    var elemIndexStr = vdboundaryContainingNode.getAttribute(VD_READABLE_ELEMENT_INDEX_ATTRIBUTE);
    var boundaryNodeIndex = parseInt(elemIndexStr);

    // now using rangy select the original boundary node element so we can calculate
    // its offset within our voicedream-node
    var rangyElem = rangy.createRange();
    rangyElem.selectNode(boundaryContainer);

    // Get the range of the selected node within its voicedream-readable
    // container node
    var rangeInContainer = rangyElem.toCharacterRange(vdboundaryContainingNode);
    var offsetInVoiceDreamNode = rangeInContainer.start + boundaryOffsetInContainer;

    return Object.create({index:boundaryNodeIndex, offset:offsetInVoiceDreamNode});

}